Package: SAIGEgds
Type: Package
Title: Scalable Implementation of Generalized mixed models using GDS
        files in Phenome-Wide Association Studies
Version: 2.12.0
Date: 2026-04-23
Depends: R (>= 4.0.0), gdsfmt (>= 1.28.0), SeqArray (>= 1.50.2), Rcpp
LinkingTo: Rcpp, RcppArmadillo, RcppParallel (>= 5.0.0)
Imports: methods, stats, utils, Matrix, RcppParallel, SKAT,
        CompQuadForm, survey
Suggests: parallel, markdown, rmarkdown, crayon, SNPRelate, RUnit,
        knitr, ggmanh, BiocGenerics
Authors@R: c(person("Xiuwen", "Zheng", role=c("aut", "cre"),
        email="xiuwen.zheng@abbvie.com", comment=c(ORCID="0000-0002-1390-0708")),
        person("Wei", "Zhou", role="ctb",
        comment="the original author of the SAIGE R package"),
        person("J. Wade", "Davis", role="ctb"))
Description: Scalable implementation of generalized mixed models with highly
        optimized C++ implementation and integration with Genomic Data Structure
        (GDS) files. It is designed for single variant tests and set-based
        aggregate tests in large-scale Phenome-wide Association Studies (PheWAS)
        with millions of variants and samples, controlling for sample structure
        and case-control imbalance. The implementation is based on the SAIGE R
        package (v0.45, Zhou et al. 2018 and Zhou et al. 2020), and it is
        extended to include the state-of-the-art ACAT-O set-based tests.
        Benchmarks show that SAIGEgds is significantly faster than the SAIGE R
        package. Optional OpenCL-based GPU acceleration is supported for the
        GRM cross-product computation in null model fitting and for GRM
        construction.
License: GPL-3
SystemRequirements: GNU make
VignetteBuilder: knitr
ByteCompile: TRUE
URL: https://github.com/AbbVie-ComputationalGenomics/SAIGEgds
biocViews: Software, Genetics, StatisticalMethod, GenomeWideAssociation
git_url: https://git.bioconductor.org/packages/SAIGEgds
git_branch: RELEASE_3_23
git_last_commit: d1163ac
git_last_commit_date: 2026-04-28
Repository: Bioconductor 3.23
Date/Publication: 2026-04-28
NeedsCompilation: yes
Packaged: 2026-04-28 23:22:39 UTC; biocbuild
Author: Xiuwen Zheng [aut, cre] (ORCID:
    <https://orcid.org/0000-0002-1390-0708>),
  Wei Zhou [ctb] (the original author of the SAIGE R package),
  J. Wade Davis [ctb]
Maintainer: Xiuwen Zheng <xiuwen.zheng@abbvie.com>
Built: R 4.6.0; aarch64-apple-darwin23; 2026-04-29 04:58:52 UTC; unix
